<?php

/**
 * @file
 * Defines the class used for communicating with the FreshBooks server.
 *
 * Developed by Wombats - http://www.bywombats.com
 *
 * To sign up for a free trial account and support this module, use the
 * following link:
 *
 * https://bywombats.freshbooks.com/signup/
 */

// Identifies this application on the FreshBooks server.
define('FRESHBOOKS_USERAGENT', 'FreshBooks for Drupal - By Wombats');

// The string we expect when verifying credentials.
define('FRESHBOOKS_MSG_VERIFY', 'Your XML is not formatted correctly.');

// The string that indicates an API request was successful.
define('FRESHBOOKS_REQUEST_OK', 'ok');

/**
 * Defines the DrupalFreshBooks class used to communicate with the server.
 */
class DrupalFreshBooks {
  // Define variables for the API credentials.
  public $api_url = '';
  public $token = '';

  // The raw XML of the last API request and response.
  public $lastRequest = '';
  public $lastResponse = '';

  /**
   * Constructs a DrupalFreshBooks object with a set of API credentials.
   *
   * @param $fb_api_url
   *   The API URL for the FreshBooks account you want to use.
   * @param $fb_token
   *   The authentication token used to verify the connection.
   */
  function __construct($fb_api_url, $fb_token) {
    // Set the appropriate member values.
    $this->api_url = $fb_api_url;
    $this->token = $fb_token;
  }

  /**
   * Sends an API request to FreshBooks.
   *
   * @param $request
   *   The raw XML request to send to the API server.
   * @return
   *   A SimpleXMLElement object loaded with the response XML.
   */
  function sendAPIRequest($request = '<?xml version="1.0" encoding="utf-8"?>') {
    // Save the last request for later retrieval.
    $this->lastRequest = $request;

    // Initialize our cURL handler.
    $ch = curl_init();
    curl_setopt($ch, CURLOPT_URL, $this->api_url);
    curl_setopt($ch, CURLOPT_USERPWD, $this->token);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $request);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
    curl_setopt($ch, CURLOPT_TIMEOUT, 4);
    curl_setopt($ch, CURLOPT_USERAGENT, FRESHBOOKS_USERAGENT);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, TRUE);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, TRUE);

    // Retrieve a response from FreshBooks.
    $this->lastResponse = curl_exec($ch);

    // Report any errors.
    if ($error = curl_error($ch)) {
      watchdog('freshbooks', check_plain($error), array(), WATCHDOG_ERROR);
    }
    curl_close($ch);

    // Create and return a SimpleXMLElement object with the response loaded.
    try {
      $response = new SimpleXMLElement($this->lastResponse);
    }
    catch (Exception $e) {
      // Log the error message.
      watchdog('freshbooks', check_plain($e->getMessage()), array(), WATCHDOG_ERROR);

      // Create an empty response object.
      $response = new SimpleXMLElement();
    }

    return $response;
  }

  // Creates a new SimpleXMLElement object for use when building a request.
  function buildAPIRequest($method, $data) {
    // Create the barebones request XML.
    $xml = '<?xml version="1.0" encoding="utf-8"?><request method="'. $method .'"></request>';

    // Load the XML into a SimpleXMLElement object.
    $request = new SimpleXMLElement($xml);

    // Use a method specific container in the XML if need be.
    $parts = explode('.', $method);

    // Only necessary for create and update methods.
    if (in_array($parts[1], array('create', 'update'))) {
      // As of now, the API uses the first part of the method to name the
      // container.  Obviously this will need an update if they ever deviate
      // from this convention.
      $container = $request->addChild($parts[0]);
    }
    else {
      $container = $request;
    }

    // Get the acceptable arguments for the method.
    $args = $this->getMethodArgs($method);

    // Add children to the request based on the data given and the acceptable
    // arguments.
    foreach ($data as $key => $value) {
      if (in_array($key, array_keys($args))) {
        // Either add this as a child or a container if need be.
        if ($args[$key] == 'container') {
          $container2 = $container->addChild($key);

          // Add the container objects to this new container.
          foreach ($value as $key2 => $value2) {
            $container2->addChild($key2, $value2);
          }
        }
        else {
          $container->addChild($key, $value);
        }
      }
    }

    return $request;
  }

  /**
   * Returns an array of acceptable arguments and their data types for the
   *   specified API method.
   *
   * @param $method
   *   The lower case name of the FreshBooks API method to query.
   * @return
   *   An array with keys representing the arguments and values the data types.
   */
  function getMethodArgs($method) {
    switch ($method) {
      case 'client.create':
        return array('first_name' => 'str', 'last_name' => 'str', 'organization' => 'str', 'email' => 'email', 'username' => 'str', 'password' => 'str', 'work_phone' => 'phone', 'home_phone' => 'phone', 'mobile' => 'phone', 'fax' => 'phone', 'notes' => 'str', 'pstreet_1' => 'str', 'p_street2' => 'str', 'p_city' => 'str', 'p_state' => 'str', 'p_country' => 'str', 'p_code' => 'str', 's_street1' => 'str', 's_street2' => 'str', 's_city' => 'str', 's_state' => 'str', 's_country' => 'str', 's_code' => 'str');
      case 'client.update':
        return array_merge(array('client_id' => 'int'), $this->getMethodArgs('client.create'));
      case 'client.get':
        return array('client_id' => 'int');
      case 'client.delete':
        return array('client_id' => 'int');
      case 'client.list':
        return array('email' => 'str', 'username' => 'str', 'page' => 'int', 'per_page' => 'int');
      default:
        return array();
    }
  }

  // Tests the connection to verify API credentials.
  function verifyCredentials() {
    // Send an empty request to the server.
    $response = $this->sendAPIRequest();

    // We consider it a failure if the authentication failed.
    if ($response->error != FRESHBOOKS_MSG_VERIFY) {
      watchdog('freshbooks', t('Attempted to connect with invalid API credentials.'), array(), WATCHDOG_ERROR);
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Creates a new client in the FreshBooks account.
   *
   * @param $data
   *   An array of client parameters, including at least a first name, last
   *     name, organization, and e-mail address.
   * @return
   *   The client ID or FALSE if the operation failed.
   */
  function clientCreate($data) {
    // Build the request.
    $request = $this->buildAPIRequest('client.create', $data);

    // Get the response from FreshBooks.
    $response = $this->sendAPIRequest($request->asXML());

    // Return the client ID if the request was successful.
    if ($response['status'] == FRESHBOOKS_REQUEST_OK) {
      return $response->client_id;
    }
    else {
      // Otherwise log the error and return FALSE.
      watchdog('freshbooks', 'clientCreate for %first %last failed: @error', array('@first' => $data['first_name'], '@last' => $data['last_name'], '@error' => (string) $response->error), WATCHDOG_ERROR);
      return FALSE;
    }
  }

  /**
   * Updates a client in the FreshBooks account.
   *
   * @param $client_id
   *   The ID of the FreshBooks client you want to update.
   * @param $data
   *   An array of client parameters you want to update; any not specified will
   *     be left unchanged.
   * @return
   *   TRUE or FALSE indicating the success of the update.
   */
  function clientUpdate($client_id, $data = array()) {
    // Merge the client ID into the data array.
    $data = array_merge(array('client_id' => $client_id), $data);

    // Build the request.
    $request = $this->buildAPIRequest('client.update', $data);

    // Get the response from FreshBooks.
    $response = $this->sendAPIRequest($request->asXML());

    // Return TRUE if the update was successful.
    if ($response['status'] == FRESHBOOKS_REQUEST_OK) {
      return TRUE;
    }
    else {
      // Otherwise log the error and return FALSE.
      watchdog('freshbooks', 'clientUpdate for client @id failed: @error', array('@id' => $client_id, '@error' => (string) $response->error), WATCHDOG_ERROR);
      return FALSE;
    }
  }

  /**
   * Returns a particular client from the FreshBooks account.
   *
   * @param $client_id
   *   The ID of the client to retrieve.
   * @return
   *   A SimpleXMLElement object of the client if successful or FALSE if not.
   */
  function clientGet($client_id) {
    // Build the request.
    $request = $this->buildAPIRequest('client.get', array('client_id' => $client_id));

    // Get the response from FreshBooks.
    $response = $this->sendAPIRequest($request->asXML());

    // Return TRUE if the request was successful.
    if ($response['status'] == FRESHBOOKS_REQUEST_OK) {
      return $response->client;
    }
    else {
      // Otherwise log the error and return FALSE.
      watchdog('freshbooks', 'clientGet for client @id failed: @error', array('@id' => $client_id, '@error' => (string) $response->error), WATCHDOG_ERROR);
      return FALSE;
    }
  }

  /**
   * Deletes a particular client from the FreshBooks account.
   *
   * @param $client_id
   *   The ID of the client to retrieve.
   * @return
   *   TRUE or FALSE indicating the success of the delete.
   */
  function clientDelete($client_id) {
    // Build the request.
    $request = $this->buildAPIRequest('client.delete', array('client_id' => $client_id));

    // Get the response from FreshBooks.
    $response = $this->sendAPIRequest($request->asXML());

    // Return TRUE if the delete was successful.
    if ($response['status'] == FRESHBOOKS_REQUEST_OK) {
      return TRUE;
    }
    else {
      // Otherwise log the error and return FALSE.
      watchdog('freshbooks', 'clientDelete for client @id failed: @error', array('@id' => $client_id, '@error' => (string) $response->error), WATCHDOG_ERROR);
      return FALSE;
    }
  }

  /**
   * Returns a list of client summaries in order of descending client_id.
   *
   * @param $email
   *   E-mail address to use to filter the client list.
   * @param $username
   *   Username to use to filter the client list.
   * @param $page
   *   The page number to show; defaults to 1.
   * @param $per_page
   *   Number of results per page; defaults to 25.
   * @return
   *   An array of the returned SimpleXMLElement client objects if successful or
   *     an empty array on failure.
   */
  function clientList($email = NULL, $username = NULL, $page = 1, $per_page = 25) {
    // Build the data array from the parameters.
    $data = array(
      'page' => $page,
      'per_page' => $per_page,
    );

    // Only add the e-mail filter if specified.
    if (!empty($email)) {
      $data['email'] = $email;
    }

    // Only add the username filter if specified.
    if (!empty($username)) {
      $data['username'] = $email;
    }

    // Build the request.
    $request = $this->buildAPIRequest('client.list', $data);

    // Get the response from FreshBooks.
    $response = $this->sendAPIRequest($request->asXML());

    // Return the array of client objects if successful.
    if ($response['status'] == FRESHBOOKS_REQUEST_OK) {
      return $response->clients->client;
    }
    else {
      // Otherwise log the error and return FALSE.
      watchdog('freshbooks', 'clientList request failed with parameters: <pre>@data</pre>', array('@data' => $data), WATCHDOG_ERROR);
      return array();
    }
  }
}
